#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <ctype.h>


/*
format of words.tok

INDEX + WORDS

INDEX FORMAT
hi-endian 16 bit offsets for each letter
WORD0, WORD1, WORD2...   WORD25
eg:
a -> 0034
b -> 0069
c -> 0083
d -> 00AB
...
offset == 0 if letter group not used.
'a' index should always be '0x34' unless something screwy like numbers get involved.
offset begins from start of words.tok, not after the index

WORD FORMAT
BYTE	- number of characters initially used from previous word
		- this is sorta like lzw compression *I think*
		- this is 0 if the start of a new letter group.
		- this is also how the interpreter knows it's reached a new letter and to stop searching
		
BYTE/CHAR, CHAR, CHAR
character data is 7 bits, with the msb (bit 8) cleared except
for the last character of each word which has the msb SET to signal the end of the word
for encryption:
the last 7 bits are inverted, and have to be inverted (try XOR 0x7F) to get the data.

WORD - hi endian word number which is the word's group number
*/

// check for excess spaces (start, end, middle)  -check
// remove forbidden characters   -check
// change separators to spaces  -check
// check for duplicates   -check
// make sure all words start with a letter  -check

// make all lowercase   oh

// option to just list words,  list groups

// stats


// I discovered the sizes by looking at the data in the largest words.tok
// I had in my collection of games

// expanded word text around 10k max
#define STR_DATA_SIZE	30000
// never has been more than 1100 words I've found
#define WORD_LIST_SIZE	3000
// largest is 8k
#define WORDS_TOK_BUF_SIZE 30000

// excess seperator characters
#define F_EXCESS_SEP	1
// upper case letter converted to lower
#define F_UPPER			2
// illegal chracters removed
#define F_ILLEGAL_CH	8
// this is a duplicate
#define F_DUPE			16
// does not start with letter
#define F_NONALPHA	32
// ignore
#define F_DISABLE		64

char *char_separators = " ,.?!();:[]{}";
char *char_illegal = "'`-\"";	// 0x27 is '

struct sWord
{
	char *str;
	int group;
	int flags;
};

char strData[STR_DATA_SIZE];
int strDataNext = 0;

struct sWord wordList[WORD_LIST_SIZE];
int wordListNext = 0;

unsigned char wordsTok[WORDS_TOK_BUF_SIZE];
int wordsTokSize = 0;

// cleaning aspects
// ----------------------------------

// cleans the word.. separates good words by ' '
void wordClean(struct sWord *word)
{
	char buffer[200];	// buffer of previous words
	char *bufPtr;
	char *str;
	
	str = word->str;
	bufPtr = buffer;

	while (*str)
	{
		// skip excess separators at start and inbetween words
		if ( (strchr(char_separators, *str) != 0) ||
			(strchr(char_illegal, *str) != 0) )
		{
			str++;
			word->flags |= F_EXCESS_SEP;
		}
		else
		{
			assert(*str);
			do
			{
				if (strchr(char_separators, *str) != 0)
				{
					*(bufPtr++) = 0x20;	// space
					str++;
					break;
				}
				
				// if not an illegal character add to buffer
				if (strchr(char_illegal, *str) == 0)
				{
					if (isupper(*str))
						word->flags |= F_UPPER;
					*(bufPtr++) = tolower(*str);
				}
				else
					word->flags |= F_ILLEGAL_CH;
				str++;
			} while (*str != 0);
		}
	}
	
	if ((bufPtr > buffer) &&	// if buffer has been modified
		(*(bufPtr-1) == 0x20))		// and the last character is a space
	{
		bufPtr--;							// then remove space
		word->flags |= F_EXCESS_SEP;
	}
	*bufPtr = 0;
	
	// display stats
	if (word->flags & (F_ILLEGAL_CH|F_EXCESS_SEP|F_UPPER))
	{
		printf("\"%s\"[%d] cleaned to \"%s\" due to ", word->str, word->group, buffer);
		if (word->flags & F_ILLEGAL_CH)
			printf("illegal char, ");
		if (word->flags & F_UPPER)
			printf("uppercase char, ");
		if (word->flags & F_EXCESS_SEP)
			printf("excess sep char");
		printf("\n");
	}
	
	strcpy(word->str, buffer);	// overwrite.,  ignore excess removed characters
}


void wordListDupes()
{
	int cur, i;
	
	assert(wordListNext > 0);
	
	cur = 0;
	
	for (i=1; i<wordListNext; i++)
	{
		if ( !(wordList[i].flags&F_DISABLE) && 
			strcmp(wordList[i].str, wordList[cur].str)==0 )
		{
			wordList[i].flags |= F_DISABLE;
			wordList[i].flags |= F_DUPE;
			
			printf("\"%s\"[%d] is a dupe of \"%s\"[%d]\n", 
				wordList[i].str, wordList[i].group, wordList[cur].str, wordList[cur].group);
		}
		else
			cur = i;
	}
	
}

void wordListAlpha()
{
	int i;
	
	for (i=0; i<wordListNext; i++)
	{
		if (!(wordList[i].flags&F_DISABLE))
		{
			if (!isalpha(wordList[i].str[0]))
			{
				wordList[i].flags |= F_NONALPHA;
				wordList[i].flags |= F_DISABLE;
				
				printf("\"%s\"[%d] does not start with a letter\n", wordList[i].str, wordList[i].group);
			}
			
		}
		
	}
}

// used by qsort to compare against elements
int wordCompare(const void *elem1, const void *elem2)
{
	struct sWord *a, *b;
		
	a = (struct sWord *)elem1;
	b = (struct sWord *)elem2;
		
	return stricmp(a->str, b->str);
}

// sort list alphabetically
void wordListSort()
{
	qsort(wordList, wordListNext, sizeof(struct sWord), wordCompare);	
}

// display the words in list
void wordListClean()
{
	int i;
	
	// remove crap
	for (i=0; i<wordListNext; i++)
		wordClean(wordList + i);
	
	// make sure word starts with a letter
	wordListAlpha();
	
	// sort
	wordListSort();
	
	// remove dupes (requires sorting)
	wordListDupes();
}

// -----------------

// add to str data
char *strDataAdd(char *newStr)
{
	int len;
	char *strEntered;
	
	len = strlen(newStr) + 1;
	assert(len < (STR_DATA_SIZE-strDataNext) );
	strEntered = strData+strDataNext;
	strcpy(strEntered, newStr);
	strDataNext += len;  // +1 for null
	return strEntered;
}


// add to word list
void wordListAdd(char *str, int group)
{
	assert(wordListNext < WORD_LIST_SIZE);
	
	wordList[wordListNext].str = strDataAdd(str);
	wordList[wordListNext].group = group;
	wordList[wordListNext].flags = 0;
	
	wordListNext++;
}

// display the words in list
void wordListDisplay()
{
	int i;
	for (i=0; i<wordListNext; i++)
		printf("%s - %d\n", wordList[i].str, wordList[i].group);
}

void wordListStats()
{
	printf("number of words = %d\n", wordListNext);
	printf("string data size = %d\n", strDataNext);
}




unsigned char *wordsTokLoad(char *fileName)
{
	fpos_t fileSize;
	int readReturn;
	FILE *fileStream;
	
	fileStream=fopen(fileName, "rb");
	assert(fileStream);

	fseek(fileStream, 0, SEEK_END);
	fgetpos(fileStream, &fileSize);
	fseek(fileStream, 0, SEEK_SET);
	wordsTokSize = (int)fileSize;
	
	readReturn = fread(wordsTok, sizeof(unsigned char), fileSize, fileStream);
	assert(readReturn == fileSize);

	fclose(fileStream);
	
	return wordsTok;	
}

void wordsTokWrite(char *fileName)
{
	FILE *fileStream;
	int readReturn;
	
	fileStream = fopen(fileName, "wb");
	
	readReturn = fwrite(wordsTok, wordsTokSize, 1, fileStream);
	assert(readReturn == 1);
	
	fclose(fileStream);
}

void wordsTokGen()
{
	char buffer[100];	// buffer of previous words
	int bufPtr = 0;
	int i;
	unsigned char *tokPtr=0;
	int prevCount;
	unsigned char *str;
	
	tokPtr = wordsTok;
	memset(tokPtr, 0, 2*26);	// reset index to 0
	tokPtr += 2*26;	// skip past the index for the moment
	
	for (i=0; i<wordListNext; i++)
	{
		if (wordList[i].flags & F_DISABLE)
			continue;	// go to next word
		
		prevCount = 0;
		str = wordList[i].str;
		assert(str);
		
		// find out how much was in the previous word
		while ((*str!=0) && (buffer[prevCount] == *str))
		{
			str++;
			prevCount++;
		}
		strcpy(buffer, wordList[i].str);	// reset buffer to new word
		
		if (prevCount == 0)	// we have a new letter group
		{
			// set index
			int offset = tokPtr - wordsTok;
			int index = (tolower(str[0]) - 'a')*2;
			
			wordsTok[index] = (offset >> 8) & 0xFF;
			wordsTok[index + 1] = offset & 0xFF;
		}
		
		assert(*str != 0);	// this shouldn't be 0 because we'll remove dupes
		assert(prevCount <= 0xFF);	// it's an 8 bit number.. who has these big words anyways?
		*(tokPtr++) = prevCount;
		
		// output rest of word
		while (*str)
		{
			*tokPtr = *str ^ 0xFF;	// invert bits.. this is encryption!
		
			str++;
			if (*str == 0)
				*tokPtr |= 0x80;		// set high bit if the last character
			else
				*tokPtr &= 0x7F;		// clear high bit if we've got characters to go
			
			tokPtr++;
		}
	
		// output group number
		*(tokPtr++) = (wordList[i].group & 0xFF00) >> 8;
		*(tokPtr++) = wordList[i].group & 0xFF;
	}		
	
	*(tokPtr++) = 0;		// last byte is zero.
	// why?  cause when the interpreter reaches a '0' number of previous chars, it stops searching
	// this is why you have an index
	
	wordsTokSize = tokPtr - wordsTok;
}


int main()
{

	char buffer[200];
	int bufPtr = 0;
	int wordNum;
	
	int count;
	
	
	unsigned char *wordPtr;
	unsigned char *wordFileData;
	unsigned char *wordStart;

	
	printf("AGI Dictionary Cleaner v1.1\n");
	printf("Copyright (c) 2002 Nick Sonneveld (sonneveld@hotmail.com)\n\n");
	printf("separator characters defined: <<%s>>\n", char_separators);
	printf("illegal characters defined: <<%s>>\n", char_illegal);
	printf("separators converted to ' ', illegal are removed.\n\n");
	
	count = 0;
	wordFileData = wordsTokLoad("words.tok");
	
	if (wordFileData == 0)
		return 1;
	
	wordPtr = wordFileData;
	
	//printf("index:\n");
	for (count='a'; count<='z'; count++)
	{
		wordNum = *(wordPtr++) << 8;
		wordNum |= *(wordPtr++);
		
		//printf("%c -> %04X\n", count, wordNum);
		
	}
	
	count = 26*2 +2;
	
	do
	{
		wordStart = wordPtr;
		
		bufPtr = *(wordPtr++);
		count++;
		
		do
		{
			buffer[bufPtr++] = (*(wordPtr) ^ 0x7F) & 0x7F; 
			count++;
		} while ((*(wordPtr++) & 0x80) == 0);
		buffer[bufPtr] = 0;
		
		wordNum = *(wordPtr++) << 8;
		count++;
		wordNum |= *(wordPtr++);
		count++;
		
		wordListAdd(buffer, wordNum);
	} while (count <= wordsTokSize);
	
	wordListClean();
	wordListStats();
	wordsTokGen();
	wordsTokWrite("words.tok.new");
	
	printf("\ndone!  see words.tok.new for new file\n");
	
	return 0;
}
